@@ -16,11 +16,17 @@
 #' rs$run_with_output(func, args = list())
 #' rs$call(func, args = list())
 #'
+#' rs$poll_process(timeout)
+#'
 #' rs$get_state()
 #' rs$get_running_time()
 #'
 #' rs$read()
-#' rs$close()
+#' rs$close(grace = 1000)
+#'
+#' rs$traceback()
+#' rs$debug()
+#' rs$attach()
 #' ```
 #'
 #' @section Arguments:
@@ -32,6 +38,10 @@
 #' * `func`: Function object to call in the background R process.
 #'   Please read the notes for the similar argument of [r()]
 #' * `args`: Arguments to pass to the function. Must be a list.
+#' * `timeout`: Timeout period in milliseconds.
+#' * `grace`: Grace period in milliseconds, to wait for the subprocess to
+#'   exit cleanly, after its standard input is closed. If the process is
+#'   still running after this period, it will be killed.
 #'
 #' @section Details:
 #' `r_session$new()` creates a new R background process. It can wait for the
@@ -51,6 +61,10 @@
 #' returns immediately. To check if the function is done, call the
 #' `poll_process()` method.
 #'
+#' `rs$poll_process()` polls the R session with a timeout. If the session
+#' has finished the computation, it returns with `"ready"`. If the timeout
+#' is reached, it returns with `"timeout"`.
+#'
 #' `rs$get_state()` return the state of the R session. Possible values:
 #' * `"starting"`: starting up,
 #' * `"idle"`: ready to compute,
@@ -65,9 +79,19 @@
 #' available. Events might signal that the function call has finished,
 #' or they can be progress report events.
 #'
-#' `r$close()` terminates the current computation and the R process.
+#' `rs$close()` terminates the current computation and the R process.
 #' The session object will be in `"finished"` state after this.
 #'
+#' `rs$traceback() can be used after an error in the R subprocess. It is
+#' equivalent to the [traceback()] call, but it is performed in the
+#' subprocess.
+#'
+#' `rs$debug()` is an interactive debugger to inspect the dumped frames
+#' in the subprocess, after an error. See more at [r_session_debug].
+#'
+#' `rs$attach()` is an experimental function that provides a REPL
+#' (Read-Eval-Print-Loop) to the subprocess.
+#'
 #' @name r_session
 #' @examples
 #' \dontrun{
@@ -116,6 +140,11 @@ r_session <- R6::R6Class(
     poll_process = function(timeout)
       rs_poll_process(self, private, timeout),
 
+    traceback = function()
+      rs_traceback(self, private),
+    debug = function()
+      rs_debug(self, private),
+
     attach = function()
       rs_attach(self, private),
 
@@ -168,7 +197,7 @@ rs_init <- function(self, private, super, options, wait, wait_timeout) {
 
   options$func <- options$func %||% function() { }
   options$args <- list()
-  options$load_hook <- rs__make_load_hook(options$load_hook)
+  options$load_hook <- session_load_hook(options$load_hook)
 
   options <- convert_and_check_my_args(options)
   options <- setup_context(options)
@@ -178,8 +207,9 @@ rs_init <- function(self, private, super, options, wait, wait_timeout) {
 
   with_envvar(
     options$env,
-    super$initialize(options$bin, options$real_cmdargs, stdin = "|",
-                     stdout = "|", stderr = "|", poll_connection = TRUE)
+    do.call(super$initialize, c(list(options$bin, options$real_cmdargs,
+      stdin = "|", stdout = "|", stderr = "|", poll_connection = TRUE),
+      options$extra))
   )
 
   ## Make child report back when ready
@@ -209,7 +239,7 @@ rs_init <- function(self, private, super, options, wait, wait_timeout) {
     } else if (pr["process"] != "ready") {
       cat("stdout:]\n", out, "\n")
       cat("stderr:]\n", err, "\n")
-      stop("Could not start R session, timed out")
+      throw(new_error("Could not start R session, timed out"))
     }
   }
 
@@ -237,7 +267,7 @@ rs_close <- function(self, private, grace) {
   self$poll_process(grace)
   self$kill()
   self$wait(1000)
-  if (self$is_alive()) stop("Could not kill background R session")
+  if (self$is_alive()) throw(new_error("Could not kill background R session"))
   private$state <- "finished"
   private$fun_started_at <- as.POSIXct(NA)
   processx::processx_conn_close(private$pipe)
@@ -250,9 +280,9 @@ rs_call <- function(self, private, func, args) {
   ## We only allow a new command if the R session is idle.
   ## This allows keeping a clean state
   ## TODO: do we need a state at all?
-  if (private$state == "starting") stop("R session not ready yet")
-  if (private$state == "finished") stop("R session finished")
-  if (private$state == "busy") stop("R session busy")
+  if (private$state == "starting") throw(new_error("R session not ready yet"))
+  if (private$state == "finished") throw(new_error("R session finished"))
+  if (private$state == "busy") throw(new_error("R session busy"))
 
   ## Save the function in a file
   private$options$func <- func
@@ -342,7 +372,7 @@ rs_run <- function(self, private, func, args) {
   } else{
     res$stdout <- paste0(res$stdout, self$read_output())
     res$stderr <- paste0(res$stderr, self$read_error())
-    stop(res$error)
+    throw(res$error)
   }
 }
 
@@ -361,6 +391,78 @@ rs_poll_process <- function(self, private, timeout) {
   processx::poll(list(self$get_poll_connection()), timeout)[[1]]
 }
 
+rs_traceback <- function(self, private) {
+  ## TODO: get rid of magic number 12
+  traceback(utils::head(self$run(function() traceback()), -12))
+}
+
+rs_debug <- function(self, private) {
+  hasdump <- self$run(function() {
+    ! is.null(as.environment("tools:callr")$`__callr_data__`$.Last.dump)
+  })
+  if (!hasdump) stop("Can't find dumped frames, nothing to debug")
+
+  help <- function() {
+    cat("Debugging in process ", self$get_pid(),
+        ", press CTRL+C (ESC) to quit. Commands:\n", sep = "")
+    cat("  .where       -- print stack trace\n",
+        "  .inspect <n> -- inspect a frame, 0 resets to .GlobalEnv\n",
+        "  .help        -- print this message\n",
+        "  <cmd>        -- run <cmd> in frame or .GlobalEnv\n\n", sep = "")
+  }
+
+  translate_cmd <- function(cmd) {
+    if (cmd == ".where") {
+      traceback(tb)
+      if (frame) cat("Inspecting frame", frame, "\n")
+      NULL
+
+    } else if (cmd == ".help") {
+      help()
+      NULL
+
+    } else if (grepl("^.inspect ", cmd)) {
+      newframe <- as.integer(strsplit(cmd, " ")[[1]][[2]])
+      if (is.na(newframe)) {
+        message("Cannot parse frame number")
+      } else {
+        frame <<- newframe
+      }
+      NULL
+
+    } else {
+      cmd
+    }
+  }
+
+  help()
+  tb <- self$traceback()
+  frame <- 0L
+
+  while (TRUE) {
+    cat("\n")
+    prompt <- paste0(
+      "RS ", self$get_pid(),
+      if (frame) paste0(" (frame ", frame, ")"), " > ")
+    cmd <- rs__attach_get_input(prompt)
+    cmd2 <- translate_cmd(cmd)
+    if (is.null(cmd2)) next
+
+    update_history(cmd)
+
+    ret <- self$run_with_output(function(cmd, frame) {
+      dump <- as.environment("tools:callr")$`__callr_data__`$.Last.dump
+      envir <- if (!frame) .GlobalEnv else dump[[frame + 12L]]
+      eval(parse(text = cmd), envir = envir)
+    }, list(cmd = cmd, frame = frame))
+    cat(ret$stdout)
+    cat(ret$stderr)
+    if (!is.null(ret$error)) print(ret$error)
+    print(ret$result)
+  }
+  invisible()
+}
+
 rs_attach <- function(self, private) {
   out <- self$get_output_connection()
   err <- self$get_error_connection()
@@ -369,6 +471,7 @@ rs_attach <- function(self, private) {
   tryCatch({
     while (TRUE) {
       cmd <- rs__attach_get_input(paste0("RS ", self$get_pid(), " > "))
+      update_history(cmd)
       private$write_for_sure(paste0(cmd, "\n"))
       private$report_back(202, "done")
       private$attach_wait()
@@ -429,7 +532,7 @@ rs__parse_msg <- function(self, private, msg) {
   }
 
   if (! s[1] %in% names(rs__parse_msg_funcs)) {
-    stop("Unknown message code: `", s[1], "`")
+    throw(new_error("Unknown message code: `", s[1], "`"))
   }
   structure(
     rs__parse_msg_funcs[[ s[1] ]](self, private, code, message),
@@ -439,7 +542,7 @@ rs__parse_msg <- function(self, private, msg) {
 rs__parse_msg_funcs <- list()
 rs__parse_msg_funcs[["200"]] <- function(self, private, code, message) {
   if (private$state != "busy") {
-    stop("Got `done` message when session is not busy")
+    throw(new_error("Got `done` message when session is not busy"))
   }
   private$state <- "idle"
 
@@ -449,7 +552,7 @@ rs__parse_msg_funcs[["200"]] <- function(self, private, code, message) {
 
 rs__parse_msg_funcs[["201"]] <- function(self, private, code, message) {
   if (private$state != "starting") {
-    stop("Session already started, invalid `starting` message")
+    throw(new_error("Session already started, invalid `starting` message"))
   }
   private$state <- "idle"
   list(code = code, message = message)
@@ -483,17 +586,13 @@ rs__parse_msg_funcs[["501"]] <- function(self, private, code, message) {
 
 rs__parse_msg_funcs[["502"]] <- rs__parse_msg_funcs[["501"]]
 
-rs__status_expr <- function(code, text = "", fd = 3) {
+rs__status_expr <- function(code, text = "", fd = 3L) {
   substitute(
     local({
+      pxlib <- as.environment("tools:callr")$`__callr_data__`$pxlib
       code_ <- code; fd_ <- fd; text_ <- text
-      con <- processx::conn_create_fd(fd_, close = FALSE)
       data <- paste0(code_, " ", text_, "\n")
-      while (1) {
-        data <- processx::conn_write(con, data)
-        if (!length(data)) break;
-        Sys.sleep(.1)
-      }
+      pxlib$write_fd(as.integer(fd), data)
     }),
     list(code = code, fd = fd, text = text)
   )
@@ -501,14 +600,18 @@ rs__status_expr <- function(code, text = "", fd = 3) {
 
 rs__prehook <- function(stdout, stderr) {
   oexpr <- if (!is.null(stdout)) substitute({
-    .__stdout__ <- processx::conn_set_stdout(
-      drop = FALSE,
-      .__ocon__ <- processx::conn_create_file(`__fn__`, write = TRUE))
+    assign(
+      ".__stdout__",
+      as.environment("tools:callr")$`__callr_data__`$pxlib$
+                                   set_stdout_file(`__fn__`),
+      envir = as.environment("tools:callr")$`__callr_data__`)
   }, list(`__fn__` = stdout))
   eexpr <- if (!is.null(stderr)) substitute({
-    .__stderr__ <- processx::conn_set_stderr(
-      drop = FALSE,
-      .__econ__ <- processx::conn_create_file(`__fn__`, write = TRUE))
+    assign(
+      ".__stderr__",
+      as.environment("tools:callr")$`__callr_data__`$pxlib$
+                                   set_stderr_file(`__fn__`),
+      envir = as.environment("tools:callr")$`__callr_data__`)
   }, list(`__fn__` = stderr))
 
   substitute({ o; e }, list(o = oexpr, e = eexpr))
@@ -516,12 +619,14 @@ rs__prehook <- function(stdout, stderr) {
 
 rs__posthook <- function(stdout, stderr) {
   oexpr <- if (!is.null(stdout)) substitute({
-      processx::conn_set_stdout(.__stdout__)
-      close(.__ocon__);
+    as.environment("tools:callr")$`__callr_data__`$
+      pxlib$set_stdout(as.environment("tools:callr")$`__callr_data__`$
+      .__stdout__)
   })
   eexpr <- if (!is.null(stderr)) substitute({
-      processx::conn_set_stderr(.__stderr__)
-      close(.__econ__);
+    as.environment("tools:callr")$`__callr_data__`$
+      pxlib$set_stderr(as.environment("tools:callr")$`__callr_data__`$
+      .__stderr__)
   })
 
   substitute({ o; e }, list(o = oexpr, e = eexpr))
@@ -563,13 +668,6 @@ rs__get_result_and_output <- function(self, private) {
   list(result = res, stdout = stdout, stderr = stderr, error = err)
 }
 
-rs__session_load_hook <- function() {
-  expr <- substitute({
-    get("conn_disable_inheritance", asNamespace("processx"))()
-    options(error = function() invokeRestart("abort"))
-  })
-}
-
 rs__handle_condition <- function(cond) {
 
   default_handler <- function(x) {
@@ -591,14 +689,6 @@ rs__handle_condition <- function(cond) {
   invisible()
 }
 
-rs__make_load_hook <- function(user_hook) {
-  hook <- rs__session_load_hook()
-  if (!is.null(user_hook)) {
-    hook <- substitute({ d; u }, list(d = hook, u = user_hook))
-  }
-  paste0(deparse(hook), "\n")
-}
-
 ## Helper functions ------------------------------------------------------
 
 #' Create options for an [r_session] object
@@ -625,6 +715,67 @@ r_session_options_default <- function() {
     user_profile = FALSE,
     env = c(TERM = "dumb"),
     supervise = FALSE,
-    load_hook = NULL
+    load_hook = NULL,
+    extra = list()
   )
 }
+
+#' Interactive debugging of persistent R sessions
+#'
+#' The `r_session$debug()` method is an interactive debugger to inspect
+#' the stack of the background process after an error.
+#'
+#' `$debug()` starts a REPL (Read-Eval-Print-Loop), that evaluates R
+#' expressions in the subprocess. It is similar to [browser()] and
+#' [debugger()] and also has some extra commands:
+#'
+#' * `.help` prints a short help message.
+#' * `.where` prints the complete stack trace of the error. (The same as
+#'   the `$traceback()` method.
+#' * `.inspect <n>` switches the "focus" to frame `<n>`. Frame 0 is the
+#'   global environment, so `.inspect 0` will switch back to that.
+#'
+#' To exit the debugger, press the usual interrupt key, i.e. `CTRL+c` or
+#' `ESC` in some GUIs.
+#'
+#' Here is an example session that uses `$debug()` (some output is omitted
+#' for brevity):
+#'
+#' ```
+#' # ----------------------------------------------------------------------
+#' > rs <- r_session$new()
+#' > rs$run(function() knitr::knit("no-such-file"))
+#' Error in rs_run(self, private, func, args) :
+#'  callr subprocess failed: cannot open the connection
+#'
+#' > rs$debug()
+#' Debugging in process 87361, press CTRL+C (ESC) to quit. Commands:
+#'   .where       -- print stack trace
+#'   .inspect <n> -- inspect a frame, 0 resets to .GlobalEnv
+#'   .help        -- print this message
+#'   <cmd>        -- run <cmd> in frame or .GlobalEnv
+#'
+#' 3: file(con, "r")
+#' 2: readLines(input2, encoding = "UTF-8", warn = FALSE)
+#' 1: knitr::knit("no-such-file") at #1
+#'
+#' RS 87361 > .inspect 1
+#'
+#' RS 87361 (frame 1) > ls()
+#'  [1] "encoding"  "envir"     "ext"       "in.file"   "input"     "input.dir"
+#'  [7] "input2"    "ocode"     "oconc"     "oenvir"    "oopts"     "optc"
+#' [13] "optk"      "otangle"   "out.purl"  "output"    "quiet"     "tangle"
+#' [19] "text"
+#'
+#' RS 87361 (frame 1) > input
+#' [1] "no-such-file"
+#'
+#' RS 87361 (frame 1) > file.exists(input)
+#' [1] FALSE
+#'
+#' RS 87361 (frame 1) > # <CTRL + C>
+#' # ----------------------------------------------------------------------
+#' ```
+#'
+#' @name r_session_debug
+NULL
